/* $Id: cd.c,v 1.5 2002/10/13 03:40:10 jaq Exp $
 *
 * Copyright (C) 1997-2001 Id Software, Inc.
 * Copyright (c) 2002 The Quakeforge Project.
 * Copyright (C) 2008 Philippe Troin <phil@fifi.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

// Quake is a trademark of Id Software, Inc., (c) 1996 Id Software, Inc. All
// rights reserved.

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/types.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include "client.h"

static qboolean	 playing     = false;
static qboolean	 wasPlaying  = false;
static qboolean	 initialized = false;
static qboolean	 enabled     = true;
static float	 cdvolume;
static byte	 remap[100];
static pid_t	 oggplayer;
static char	 musdir[1024];
static cvar_t	*cd_volume;
static cvar_t	*cd_nocd;

void CDAudio_Stop(void);

static void startogg123(const char* opt, const char* path)
{
	int i, dsz;

	playing = true;
	oggplayer = fork();
	if (oggplayer == (pid_t)-1) {
		oggplayer = 0;
		return;
	}

	if (oggplayer != (pid_t)0)
		return;

	for (i=0, dsz = getdtablesize(); i < dsz; ++i)
		close(i);
	open("/dev/null", O_RDONLY);
	open("/dev/null", O_WRONLY);
	dup(1);

	execlp("ogg123", "ogg123", opt, path, NULL);
	_exit(255);
}

void CDAudio_Play(int track, qboolean looping)
{
	char buf[1024];

	if (playing)
		CDAudio_Stop();

	snprintf(buf, sizeof(buf), "%s/track%02d.ogg", musdir, track);
	if (access(buf, R_OK) != 0)
		return;

	startogg123((looping ? "-qr" : "-q"), buf);
}

void CDAudio_RandomPlay(void)
{
	if (playing)
		CDAudio_Stop();

	startogg123("-Z", musdir);
}

void CDAudio_Stop()
{
	if (!enabled)
		return;
	if (!playing)
		return;

	kill(oggplayer, SIGTERM);
	kill(oggplayer, SIGCONT);
	waitpid(oggplayer, NULL, 0);
	oggplayer = (pid_t)0;


	wasPlaying = false;
	playing	   = false;
}

void CDAudio_Pause(void)
{
	if (!enabled)
		return;

	if (!playing)
		return;


	kill(oggplayer, SIGSTOP);

	wasPlaying = playing;
	playing = false;
}


void CDAudio_Resume(void)
{
	if (!enabled)
		return;

	if (!wasPlaying)
		return;

	kill(oggplayer, SIGCONT);

	playing = true;
}

static void CD_f (void)
{
	char	*command;
	int		ret;
	int		n;

	if (Cmd_Argc() < 2)
		return;

	command = Cmd_Argv (1);

	if (Q_strcasecmp(command, "on") == 0)
	{
		enabled = true;
		return;
	}

	if (Q_strcasecmp(command, "off") == 0)
	{
		if (playing)
			CDAudio_Stop();
		enabled = false;
		return;
	}

	if (Q_strcasecmp(command, "reset") == 0)
	{
		enabled = true;
		if (playing)
			CDAudio_Stop();
		for (n = 0; n < 100; n++)
			remap[n] = n;
		return;
	}

	if (Q_strcasecmp(command, "remap") == 0)
	{
		ret = Cmd_Argc() - 2;
		if (ret <= 0)
		{
			for (n = 1; n < 100; n++)
				if (remap[n] != n)
					Com_Printf("  %u -> %u\n", n, remap[n]);
			return;
		}
		for (n = 1; n <= ret; n++)
			remap[n] = atoi(Cmd_Argv (n+1));
		return;
	}

	if (Q_strcasecmp(command, "close") == 0)
	{
		return;
	}

	if (Q_strcasecmp(command, "play") == 0)
	{
		CDAudio_Play((byte)atoi(Cmd_Argv (2)), false);
		return;
	}

	if (Q_strcasecmp(command, "loop") == 0)
	{
		CDAudio_Play((byte)atoi(Cmd_Argv (2)), true);
		return;
	}

	if (Q_strcasecmp(command, "stop") == 0)
	{
		CDAudio_Stop();
		return;
	}

	if (Q_strcasecmp(command, "pause") == 0)
	{
		CDAudio_Pause();
		return;
	}

	if (Q_strcasecmp(command, "resume") == 0)
	{
		CDAudio_Resume();
		return;
	}

	if (Q_strcasecmp(command, "eject") == 0)
	{
		if (playing)
			CDAudio_Stop();
		return;
	}

	if (Q_strcasecmp(command, "info") == 0)
	{
		if (playing)
			Com_Printf("Currently playing\n");
		else if (wasPlaying)
			Com_Printf("Paused\n");
		Com_Printf("Volume is %f\n", cdvolume);
		return;
	}
}

void CDAudio_Update(void)
{
}

int CDAudio_Init(void)
{
	int i;
	cvar_t			*cv;
	char *path;

	if (initialized)
		return 0;

	cv = Cvar_Get ("nocdaudio", "0", CVAR_NOSET);
	if (cv->value)
		return -1;

	cd_nocd = Cvar_Get ("cd_nocd", "0", CVAR_ARCHIVE );
	if ( cd_nocd->value)
		return -1;

	cd_volume = Cvar_Get ("cd_volume", "1", CVAR_ARCHIVE);

	musdir[0] = 0;
	for (path = FS_NextPath(NULL); path; path = FS_NextPath(path))
	{
		struct stat sbuf;
		char buf[sizeof(musdir)];
		snprintf(buf, sizeof(buf), "%s/music", path);
		if (stat(buf, &sbuf) == 0 && S_ISDIR(sbuf.st_mode))
		{
			strcpy(musdir, buf);
			break;
		}
	}
	if (musdir[0] == 0)
	{
		Com_Printf("No music directory found for CD Audio\b");
		return -1;
	}

	for (i = 0; i < 100; i++)
		remap[i] = i;
	initialized = true;
	enabled = true;
	Cmd_AddCommand ("cd", CD_f);

	Com_Printf("CD Audio Initialized, music in %s\n", musdir);

	return 0;
}

void CDAudio_Activate (qboolean active)
{
	if (active)
		CDAudio_Resume ();
	else
		CDAudio_Pause ();
}

void CDAudio_Shutdown(void)
{
	if (!initialized)
		return;
	CDAudio_Stop();
	initialized = false;
}
